/*--------------------------------------------------------------------------
 *
 * 3DM-G Interface Software
 *
 * (c) 2003 Microstrain, Inc.
 * All rights reserved.
 *
 * www.microstrain.com
 * 310 Hurricane Lane, Suite 4
 * Williston, VT 05495 USA
 * Tel: 802-862-6629 tel
 * Fax: 802-863-4093 fax
 *--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------
 * m3dmgSerialWin.c
 *
 * Serial port interface for the 3DM-G Sensor device. 
 *
 * This module is specific to the Microsoft Windows platform. (95,98,NT,2000,XP)
 * For a Linux version please use m3dmgSerialLinux.c
 *
 * Note: The Win and POSIX versions are sufficiently different that we opted for
 * separate source files rather than inserting a large number of pre-processor
 * directives based on opeating system flags. 
 *
 * rev 0.1 - fixed static string problem in openPort. JHS Apr-2-2003
 *--------------------------------------------------------------------------*/

#include <stdio.h>
#include "windows.h"
#include "m3dmgSerial.h"
#include "m3dmgErrors.h"

// port Handles in use
HANDLE portHandles[MAX_PORT_NUM];

/*--------------------------------------------------------------------------
 * openPort
 *
 * parameters:  portNum : a port number from 1 to MAX_PORT_NUM
 *
 * returns:     a port handle or M3D_COMM_FAILED.
 *--------------------------------------------------------------------------*/

int openPort(int portNum) {

    HANDLE portHandle;
    char *portBase = "COM";
    char *portIndex = "0";
    char *portName;

    if (portNum<0 || portNum > MAX_PORT_NUM) {
        return M3D_COMM_INVALID_PORTNUM;
    }

    _itoa(portNum, portIndex,10);

    portName = (char *)calloc(strlen(portBase) + 
                              strlen(portIndex) + 1, sizeof(char));
    strcat(portName, portBase);
    strcat(portName, portIndex);

	portHandle = CreateFile(portName, 
							GENERIC_READ | GENERIC_WRITE,
							0, // exclusive access
							NULL, // no security
							OPEN_EXISTING,
							0, // no overlapped I/O
							NULL); // null template

    printf("trying to open port: %s # %d\n", portName, portNum);

    free(portName);

	if (portHandle == INVALID_HANDLE_VALUE) {
		return M3D_COMM_FAILED;
	} else {
        SetupComm(portHandle, 128, 128);	// set buffer sizes
        portHandles[portNum-1] = portHandle;
        return portNum;
	}
}

/*--------------------------------------------------------------------------
 * closePort
 *
 * parameters:  portNum : closes the port corresponding to this port number.
 *--------------------------------------------------------------------------*/

void closePort(int portNum) {
    if (portNum>0 && portNum<=MAX_PORT_NUM) {
    	CloseHandle(portHandles[portNum-1]);
    }
}

/*--------------------------------------------------------------------------
 * setCommParameters
 *
 * parameters:  portNum  : a serial port number
 *              baudrate : the communication flow rate (speed).
 *              charsize : the character size (7,8 bits)
 *              parity   : 0=none, 1=odd, 2=even
 *              stopbits : the number of stop bits (1 or 2)
 *
 * returns:     M3D_COMM_OK if the settings succeeded.
 *              M3D_COMM_FAILED if there was an error.
 *--------------------------------------------------------------------------*/

int setCommParameters(int portNum,
                      int baudrate,
                      int charsize,
                      int parity,
                      int stopbits ) {

    BOOL ready;
	DCB  dcb;
    HANDLE portHandle;
    portHandle = portHandles[portNum-1];

	ready = GetCommState(portHandle, &dcb);
    if (!ready) {
        return M3D_COMM_FAILED;
    }
    if (stopbits==1) {
        stopbits=0;    /* in Windows 0 means 1 stopbit */
    }
	dcb.BaudRate = baudrate;
	dcb.ByteSize = charsize;
	dcb.Parity = parity;
	dcb.StopBits = stopbits;
	dcb.fAbortOnError = TRUE;
	ready = SetCommState(portHandle, &dcb);
    if (!ready)
        return M3D_COMM_FAILED;
    else
        return M3D_COMM_OK;
}

/*--------------------------------------------------------------------------
 * setCommTimeouts
 *
 * parameters:  portNum      : a serial port number
 *              readTimeOut  : timeout for single char and total read.
 *              writeTimeOut : timeout for single char and total write.
 *--------------------------------------------------------------------------*/

int setCommTimeouts( int portNum, 
                     int readTimeout, 
                     int writeTimeout) {
    BOOL ready;
	COMMTIMEOUTS timeOuts;
    HANDLE portHandle;
    portHandle = portHandles[portNum-1];

	ready = GetCommTimeouts (portHandle, &timeOuts);
    if (!ready) {
        return M3D_COMM_FAILED;
    }
	timeOuts.ReadIntervalTimeout = readTimeout;
	timeOuts.ReadTotalTimeoutConstant = readTimeout;
	timeOuts.ReadTotalTimeoutMultiplier = 10;
	timeOuts.WriteTotalTimeoutConstant = writeTimeout;
	timeOuts.WriteTotalTimeoutMultiplier = 10;
	ready = SetCommTimeouts (portHandle, &timeOuts);
    if (!ready) {
        return M3D_COMM_FAILED;
    }
    else
        return M3D_COMM_OK;
}

/*--------------------------------------------------------------------------
 * sendData
 *
* parameters:  portNum  - a serial port number (1..n)
 *              command  - a single byte command
 *              commandLength - the length of the command buffer
 *                              which is the number of bytes to send
 *
 * returns:     COMM_OK if write/read succeeded
 *              COMM_WRITE_ERROR if there was an error writing to the port
 *--------------------------------------------------------------------------*/

int sendData(int portNum, char* command, int commandLength)
{
	DWORD bytesWritten;
	BOOL status;

    HANDLE portHandle;
    portHandle = portHandles[portNum-1];

    /* write the command. */
	status = WriteFile (portHandle, command, commandLength, &bytesWritten, 0);
	if (!status)
		return M3D_COMM_WRITE_ERROR;  //!! check bytes written?
    else
        return M3D_COMM_OK;
}

/*--------------------------------------------------------------------------
 * receiveData
 *
 * parameters:  portNum  - a serial port number
 *              response - a pointer to a character buffer to hold the response
 *              responseLength - the # of bytes of the expected response.
 *
 * returns:     COMM_OK if write/read succeeded
 *              COMM_READ_ERROR if there was an reading from the port
 *              COMM_RLDLEN_ERROR if the length of the response did not match
 *                              the number of returned bytes.
 *--------------------------------------------------------------------------*/

int receiveData(int portNum, char *response, int responseLength) {
	DWORD bytesRead;
	BOOL status;
    int returnVal;

    HANDLE portHandle;
    portHandle = portHandles[portNum-1];

	/* receive response data from the serial port if expected. */
    returnVal = M3D_COMM_OK;
	if (responseLength>0) {
		status = ReadFile(portHandle, &response[0], responseLength, &bytesRead, NULL);
		if (status) {
            /* check for wrong number of bytes returned. */
            if (bytesRead != responseLength)
                returnVal = M3D_COMM_RDLEN_ERROR;

		} else
            returnVal = M3D_COMM_READ_ERROR;
	}
    return returnVal;
}

/*-------------- end of m3dmgSerialWin.c ----------------------*/
