/*--------------------------------------------------------------------------
 *
 * 3DM-G Interface Software
 *
 * (c) 2003 Microstrain, Inc.
 * All rights reserved.
 *
 * www.microstrain.com
 * 310 Hurricane Lane, Suite 4
 * Williston, VT 05495 USA
 * Tel: 802-862-6629 
 * Fax: 802-863-4093
 *--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------
 * m3dmgloop.c
 *
 * A test application which reads a transform matrix in a loop.
 * We measure how long it takes to perform a certain number of 
 * successive reads.
 *
 *  Win32 platforms:
 *      cl m3dmgloop.c m3dmgAdapter.c m3dmgErrors.c m3dmgSerialWin.c -o m3dmgloop
 *
 *  Linux platform (with gcc) :
 *      gcc -DLINUX_OS m3dmgloop.c m3dmgAdapter.c m3dmgErrors.c m3dmgSerialLinux.c -o m3dmgloop
 *--------------------------------------------------------------------------*/

#include <stdio.h>
#include <sys/types.h>
#include <time.h>

#include "m3dmgErrors.h"
#include "m3dmgSerial.h"
#include "m3dmgAdapter.h"

int main(int argc, char **argv) {

    int portNum, deviceNum;
    int i, j;

/* variables used to retieve and store sensor data */
    int sn;
    char *fw;
    char *axis[] = { "X", "Y", "Z"};
    float quart[3];     /*  quarternions */
    float xform[3][3];  /*  transformation matrix */

    int errorCode;
    int loopCount, failCount;
    int tryPortNum = 1;
    time_t tstart, tend;

	printf("3DM-G Iteration Test\n");

    if (argc > 1) {
        tryPortNum = atoi(argv[1]);
        if (tryPortNum < 1 || tryPortNum > 9) {
            printf("usage:  m3dmgloop <portNumber>\n");
            printf("        valid ports are 1..9, default is 1\n");
            exit(1);
        }
    }


    /* open a port, map a device */
    portNum = m3dmg_openPort(tryPortNum, 38400, 8, 0, 1);
    if (portNum<0) {
        printf("port open failed.\n");
        printf("Comm error %d, %s: ", portNum, explainError(portNum));
        return -1;
    }
    printf("Port is open as #%d - OK.\n", portNum);

    /* map device - this is required! */
    deviceNum = m3dmg_mapDevice(1, portNum);
    if (deviceNum <= 0) {
        printf("Could not map the Device to a Port\n");
        printf("%s\n", explainError(deviceNum));
        return -1;
    }
    printf("Device Number is #: %d\n\n", deviceNum);

    /* get serial number */
    errorCode = m3dmg_getSerialNumber(deviceNum, &sn);
    if (errorCode != M3D_OK)
        printf("S/N Error - : %s\n", explainError(errorCode));
    else
        printf("Serial #: %d\n", sn);
    printf("\n");

    /* get firmware number (as a string) */
    fw = (char *) malloc( (size_t)12 );
    errorCode = m3dmg_getFirmwareVersion(deviceNum, fw);
    printf("Firmware ");
    if (errorCode < 0)
        printf("Error - : %s\n", explainError(errorCode));
    else
        printf("#: %s\n", fw);
    printf("\n");

    failCount = 0;
    loopCount = 100;
    printf("Instantaneous orientation matrix - read %d times\n", loopCount);
    time(&tstart);
    for (i=0; i< loopCount; i++) {
        /* instantaneous orientation matrix */
        errorCode = m3dmg_getOrientMatrix(deviceNum, &xform[0], M3D_INSTANT);
        if (errorCode < 0)
            failCount++;
    }
    time(&tend);
    printf("Time to complete %d reads: %d\n", loopCount, tend-tstart);
    printf("Average time per read %3.4f sec\n", (float) (tend-tstart)/loopCount);

    if (failCount == 0)
        printf("No failures.\n");
    else
        printf("Failed %d times.\n", failCount);

    /* free allocated memory */
    free(fw);

    /* close device */
    m3dmg_closeDevice(deviceNum);
	return(0);
}

/*-------------- end of m3dmgloop.c ----------------------*/
